import logging
import os

import ckan.plugins as p
from ckan import model
from ckan.lib.plugins import DefaultOrganizationForm

from ckanext.hierarchy.logic import action
from ckanext.hierarchy import helpers

log = logging.getLogger(__name__)
g = p.toolkit.g

# This plugin is designed to work only these versions of CKAN
p.toolkit.check_ckan_version(min_version='2.0')


def custom_convert_from_extras(key, data, errors, context):
    '''Converts values from extras, tailored for groups.'''

    # Set to empty string to remove Missing objects
    data[key] = ""

    to_remove = []
    for data_key in list(data.keys()):
        if (data_key[0] == 'extras'):
            data_value = data[data_key]
            if 'key' in data_value and data_value['key'] == key[-1]:
                data[key] = data_value['value']
                to_remove.append(data_key)
                break
    else:
        return

    for remove_key in to_remove:
        del data[remove_key]


class HierarchyDisplay(p.SingletonPlugin):
    p.implements(p.IConfigurer, inherit=True)
    p.implements(p.IActions, inherit=True)
    p.implements(p.ITemplateHelpers, inherit=True)
    p.implements(p.IPackageController, inherit=True)

    # IConfigurer

    def update_config(self, config):
        p.toolkit.add_template_directory(config, 'templates')
        p.toolkit.add_public_directory(config, 'fanstatic')
        p.toolkit.add_resource('fanstatic', 'hierarchy')

        try:
            from ckan.lib.webassets_tools import add_public_path
        except ImportError:
            pass
        else:
            asset_path = os.path.join(
                os.path.dirname(__file__), 'fanstatic'
            )
            add_public_path(asset_path, '/')

    # IActions

    def get_actions(self):
        return {'group_tree': action.group_tree,
                'group_tree_section': action.group_tree_section,
                }

    # ITemplateHelpers
    def get_helpers(self):
        return {'group_tree': helpers.group_tree,
                'group_tree_section': helpers.group_tree_section,
                'group_tree_parents': helpers.group_tree_parents,
                'group_tree_get_longname': helpers.group_tree_get_longname,
                'group_tree_highlight': helpers.group_tree_highlight,
                'get_allowable_parent_groups':
                helpers.get_allowable_parent_groups,
                'is_include_children_selected':
                helpers.is_include_children_selected,
                }

    # IPackageController

    def before_dataset_search(self, search_params):
        '''When searching an organization, optionally extend the search any
        sub-organizations too. This is achieved by modifying the search options
        before they go to SOLR.
        '''
        # Check if we're called from the organization controller, as detected
        # by g being registered for this thread, and the existence of g.fields
        # values

        try:
            if not isinstance(g.fields, list) and not hasattr(g, 'fields'):
                return search_params
        except (TypeError, AttributeError, RuntimeError):
            # it's non-organization controller or CLI call
            return search_params

        # e.g. search_params['q'] = u' owner_org:"id" include_children: "True"'
        query = search_params.get('q', '')
        fq = search_params.get('fq', '')

        include_children = query and 'include_children: "True"' in query

        if include_children or helpers.is_include_children_selected():

            # get a list of all the children organizations and include them in
            # the search params
            children_org_hierarchy = model.Group.get(g.group_dict.get('id')).\
                get_children_group_hierarchy(type='organization')
            children_names = [org[1] for org in children_org_hierarchy]

            # remove include_children clause - it is a message for this func,
            # not solr
            # CKAN<=2.7 it's in the q field:
            query = query.replace('include_children: "True"', '')
            # CKAN=2.8.x it's in the fq field:
            fq = fq.replace('include_children:"True"', '')

            if children_names:
                # remove existing owner_org:"<parent>" clause - we'll replace
                # it with the tree of orgs in a moment.
                owner_org_q = 'owner_org:"{}"'.format(g.group_dict.get('id'))
                # CKAN<=2.7 it's in the q field:
                query = query.replace(owner_org_q, '')
                # CKAN=2.8.x it's in the fq field:
                fq = fq.replace(owner_org_q, '')
                # add the org clause
                query = query.strip()
                if query:
                    query += ' AND '
                query += '({})'.format(
                    ' OR '.join(
                        'organization:{}'.format(org_name)
                        for org_name in [g.group_dict.get('name')] +
                        children_names))

            search_params['q'] = query.strip()
            search_params['fq'] = fq
            # remove include_children from the filter-list - we have a checkbox
            g.fields_grouped.pop('include_children', None)

        return search_params

    before_search = before_dataset_search


class HierarchyForm(p.SingletonPlugin, DefaultOrganizationForm):
    p.implements(p.IGroupForm, inherit=True)

    # IGroupForm

    def group_types(self):
        return ('organization',)

    def group_controller(self):
        return 'organization'

    def setup_template_variables(self, context, data_dict):
        group_id = data_dict.get('id')
        g.allowable_parent_groups = \
            helpers.get_allowable_parent_groups(group_id)



'''Tests that this extension doesn't break unrelated things'''

import pytest
from ckan.tests import helpers


@pytest.mark.usefixtures('clean_db', 'clean_index', 'with_request_context')
class TestSearchApi():

    def test_package_search_is_unaffected(self, initial_data, app):
        parent_org, child_org, parent_dataset, child_dataset = \
            initial_data

        # package_search API is unaffected by ckanext-hierarchy (only searches
        # via the front-end are affected)
        package_search_result = helpers.call_action(
            'package_search',
            fq='owner_org:{}'.format(parent_org['id']))
        search_results = \
            [result['name'] for result in package_search_result['results']]

        assert set(search_results) == set(('parent',))


@pytest.mark.usefixtures('clean_db', 'clean_index', 'with_request_context')
class TestPages():

    def test_home_page(self, initial_data, app):
        parent_org, child_org, parent_dataset, child_dataset = \
            initial_data

        app.get(url='/')



from bisect import bisect_right


class GroupTreeNode(dict):
    '''Represents a group in a tree, used when rendering the tree.

    Is a dict, with links to child GroupTreeNodes, so that it is already
    'dictized' for output from the logic layer.
    '''
    def __init__(self, group_dict):
        dict.__init__(self)
        self.update(group_dict)
        self['highlighted'] = False
        self['children'] = []
        # self._children_titles has a 1:1 relationship with values in
        # self.['children'], and is used to help keep them sorted by title
        self._children_titles = []

    def add_child_node(self, child_node):
        '''Adds the child GroupTreeNode to this node, keeping the children
        in alphabetical order by title.
        '''
        title = child_node['title']
        insert_index = bisect_right(self._children_titles, title)
        self['children'].insert(insert_index, child_node)
        self._children_titles.insert(insert_index, title)

    def highlight(self):
        '''Flag this group to indicate it should be shown highlighted
        when rendered.'''
        self['highlighted'] = True


def group_dictize(group):
    '''Convert a Group object into a dict suitable for GroupTreeNode
    Much simpler and quicker to do this than run than the full package_show.
    '''
    return {'id': group.id,
            'name': group.name,
            'title': group.title,
            'type': group.type}



import ckan.plugins as p
import ckan.model as model
from ckan.common import request, is_flask_request


def group_tree(organizations=[], type_='organization'):
    full_tree_list = p.toolkit.get_action('group_tree')({}, {'type': type_})

    if not organizations:
        return full_tree_list
    else:
        filtered_tree_list = group_tree_filter(organizations, full_tree_list)
        return filtered_tree_list


def group_tree_filter(organizations, group_tree_list, highlight=False):
    # this method leaves only the sections of the tree corresponding to the
    # list since it was developed for the users, all children organizations
    # from the organizations in the list are included
    def traverse_select_highlighted(group_tree, selection=[], highlight=False):
        # add highlighted branches to the filtered tree
        if group_tree['highlighted']:
            # add to the selection and remove highlighting if necessary
            if highlight:
                selection += [group_tree]
            else:
                selection += group_tree_highlight([], [group_tree])
        else:
            # check if there is any highlighted child tree
            for child in group_tree.get('children', []):
                traverse_select_highlighted(child, selection)

    filtered_tree = []
    # first highlights all the organizations from the list in the three
    for group in group_tree_highlight(organizations, group_tree_list):
        traverse_select_highlighted(group, filtered_tree, highlight)

    return filtered_tree


def group_tree_section(id_, type_='organization', include_parents=True,
                       include_siblings=True):
    return p.toolkit.get_action('group_tree_section')(
        {'include_parents': include_parents,
         'include_siblings': include_siblings},
        {'id': id_, 'type': type_, })


def group_tree_parents(id_, type_='organization'):
    tree_node = p.toolkit.get_action('organization_show')({}, {'id': id_,
                                                               'include_dataset_count': False,
                                                               'include_users': False,
                                                               'include_followers': False,
                                                               'include_tags': False})
    if (tree_node['groups']):
        parent_id = tree_node['groups'][0]['name']
        parent_node = \
            p.toolkit.get_action('organization_show')({}, {'id': parent_id})
        return group_tree_parents(parent_id) + [parent_node]
    else:
        return []


def group_tree_get_longname(id_, default="", type_='organization'):
    tree_node = p.toolkit.get_action('organization_show')({}, {'id': id_,
                                                               'include_dataset_count': False,
                                                               'include_users': False,
                                                               'include_followers': False,
                                                               'include_tags': False})
    longname = tree_node.get("longname", default)
    if not longname:
        return default
    return longname


def group_tree_highlight(organizations, group_tree_list):
    def traverse_highlight(group_tree, name_list):
        if group_tree.get('name', "") in name_list:
            group_tree['highlighted'] = True
        else:
            group_tree['highlighted'] = False
        for child in group_tree.get('children', []):
            traverse_highlight(child, name_list)

    selected_names = [o.get('name', None) for o in organizations]

    for group in group_tree_list:
        traverse_highlight(group, selected_names)
    return group_tree_list


def get_allowable_parent_groups(group_id):
    if group_id:
        group = model.Group.get(group_id)
        allowable_parent_groups = \
            group.groups_allowed_to_be_its_parent(type='organization')
    else:
        allowable_parent_groups = model.Group.all(
            group_type='organization')
    return allowable_parent_groups


def is_include_children_selected():
    include_children_selected = False
    if is_flask_request():
        if request.params.get('include_children'):
            include_children_selected = True
    return include_children_selected



import pytest
from bs4 import BeautifulSoup


@pytest.mark.usefixtures('clean_db', 'clean_index', 'with_request_context')
class TestOrgPage():

    def test_search_parent_including_children(self, initial_data, app):
        parent_org, child_org, parent_dataset, child_dataset = \
            initial_data

        response = app.get(
            url='/organization/parent_org?include_children=True')

        search_results = scrape_search_results(response)
        assert search_results == set(('parent', 'child'))

    def test_search_parent_excluding_children(self, initial_data, app):
        parent_org, child_org, parent_dataset, child_dataset = \
            initial_data

        response = app.get(
            url='/organization/parent_org')

        search_results = scrape_search_results(response)
        assert search_results == set(('parent',))

    def test_search_child_including_children(self, initial_data, app):
        parent_org, child_org, parent_dataset, child_dataset = \
            initial_data

        response = app.get(
            url='/organization/child_org?include_children=True')

        search_results = scrape_search_results(response)
        assert search_results == set(('child',))


def scrape_search_results(response):
    try:
        body = response.data
    except AttributeError:
        body = response.body
    soup = BeautifulSoup(body)
    dataset_names = set()
    for dataset_li in soup.find_all('li', class_='dataset-item'):
        dataset_names.add(dataset_li.find('a')['href'].split('/')[-1])
    return dataset_names



import pytest

from ckan.tests import factories
from ckan import model


@pytest.fixture
def initial_data():
    parent_org = factories.Organization(name='parent_org', title='Parent')
    child_org = factories.Organization(name='child_org', title='child')
    member = model.Member(
        group=model.Group.get(child_org['id']),
        table_id=parent_org['id'], table_name='group', capacity='parent')
    model.Session.add(member)
    model.Session.commit()
    parent_dataset = factories.Dataset(name='parent', title='Parent',
                                       owner_org=parent_org['id'])
    child_dataset = factories.Dataset(name='child', title='Child',
                                      owner_org=child_org['id'])
    factories.Dataset(name='decoy', title='Decoy')
    return parent_org, child_org, parent_dataset, child_dataset



import logging

import ckan.plugins as p
import ckan.logic as logic
from ckanext.hierarchy.model import GroupTreeNode

log = logging.getLogger(__name__)
_get_or_bust = logic.get_or_bust


@logic.side_effect_free
def group_tree(context, data_dict):
    '''Returns the full group tree hierarchy.

    :returns: list of top-level GroupTreeNodes
    '''
    model = _get_or_bust(context, 'model')
    group_type = data_dict.get('type', 'group')
    return [_group_tree_branch(group, type=group_type)
            for group in model.Group.get_top_level_groups(type=group_type)]


@logic.side_effect_free
def group_tree_section(context, data_dict):
    '''Returns the section of the group tree hierarchy which includes the given
    group, from the top-level group downwards.

    :param id: the id or name of the group to include in the tree
    :param include_parents: if false, starts from given group
    :param include_siblingss: if false, excludes given group siblings
    :returns: the top GroupTreeNode of the tree section
    '''
    group_name_or_id = _get_or_bust(data_dict, 'id')
    model = _get_or_bust(context, 'model')
    group = model.Group.get(group_name_or_id)
    if group is None:
        raise p.toolkit.ObjectNotFound
    group_type = data_dict.get('type', 'group')
    if group.type != group_type:
        how_type_was_set = 'was specified' if data_dict.get('type') \
            else 'is filtered by default'
        raise p.toolkit.ValidationError(
            'Group type is "%s" not "%s" that %s' %
            (group.type, group_type, how_type_was_set))
    include_parents = context.get('include_parents', True)
    include_siblings = context.get('include_siblings', True)
    if include_parents:
        root_group = (group.get_parent_group_hierarchy(type=group_type)
                      or [group])[0]
    else:
        root_group = group
    if include_siblings or root_group == group:
        return _group_tree_branch(root_group, highlight_group_name=group.name,
                                  type=group_type)
    else:
        section_subtree = _group_tree_branch(group,
                                             highlight_group_name=group.name,
                                             type=group_type)
        return _nest_group_tree_list(
            group.get_parent_group_hierarchy(type=group_type),
            section_subtree)


def _nest_group_tree_list(group_tree_list, group_tree_leaf):
    '''Returns a tree branch composed by nesting the groups in the list.

    :param group_tree_list: list of groups to build a tree, first is root
    :returns: the top GroupTreeNode of the tree
    '''
    root_node = None
    last_node = None
    for group in group_tree_list:
        node = GroupTreeNode(
            {'id': group.id,
             'name': group.name,
             'title': group.title})
        if not root_node:
            root_node = last_node = node
        else:
            last_node.add_child_node(node)
            last_node = node
    last_node.add_child_node(group_tree_leaf)
    return root_node


def _group_tree_branch(root_group, highlight_group_name=None, type='group'):
    '''Returns a branch of the group tree hierarchy, rooted in the given group.

    :param root_group_id: group object at the top of the part of the tree
    :param highlight_group_name: group name that is to be flagged 'highlighted'
    :returns: the top GroupTreeNode of the tree
    '''
    nodes = {}  # group_id: GroupTreeNode()
    root_node = nodes[root_group.id] = GroupTreeNode(
        {'id': root_group.id,
         'name': root_group.name,
         'title': root_group.title})
    if root_group.name == highlight_group_name:
        nodes[root_group.id].highlight()
        highlight_group_name = None
    for group_id, group_name, group_title, parent_id in \
            root_group.get_children_group_hierarchy(type=type):
        node = GroupTreeNode({'id': group_id,
                              'name': group_name,
                              'title': group_title})
        nodes[parent_id].add_child_node(node)
        if highlight_group_name and group_name == highlight_group_name:
            node.highlight()
        nodes[group_id] = node
    return root_node
